from enum import Enum

class Effect(Enum):
    """Generic effects which may be interpreted differently or not at all
    by different devices"""
    NONE = "None"
    JUMP_7 = "Jump 7"
    JUMP_3 = "Jump 3"
    FADE_7 = "Fade 7"
    FADE_3 = "Fade 3"
    FLASH  = "Flash"
    AUTO   = "Auto"
    # Slowly fades through colors. Noise increases speed.
    FADE_7_FAST_ON_NOISE = "Fade 7 Volume Controls Speed"
    # Slowly pulses each color twice. Noise increases speed.
    TWO_FADE_FAST_ON_NOISE = "Two Fade Volume Controls Speed"
    # Solid color. Noise quickly cycles through colors.
    JUMP_ON_NOISE_PAUSE_QUIET = "Cycle Colors When Loud"
    # Off. Noise turns on lights and quickly cycles through colors.
    JUMP_ON_NOISE_OFF_QUIET = "Light and Cycle Colors When Loud"


from dataclasses import dataclass
from enum import Enum
from typing import Callable

from bleak import BleakClient
from bleak.backends.device import BLEDevice
from cheshire.communication.bluetooth import BLETransmitter, GattProfile
from cheshire.communication.transmitter import Transmitter
from cheshire.compiler.compiler import StateCompiler
from cheshire.compiler.state import LightState
from cheshire.generic.command import *

from cheshire.generic.command import Command
from .compilers.ks03_old.compiler import KS03OldCompiler
from .compilers.ks03_new.compiler import KS03NewCompiler

class DeviceNamePrefix(Enum):
    """
    Device types are determined from the Bluetooth device name prefix.
    """
    KS01 = "KS01-"
    KS02 = "KS02-"
    KS03 = "KS03-"
    KS03_New = "KS03~"
    KS04 = "KS04-"
    KS04_New = "KS04~"
    KS05 = "KS05-"
    KS07 = "KS07-"
    KS08 = "KS08-"
    KS09 = "KS09-"
    KS10 = "KS10-"
    KS11 = "KS11-" # Seems to have white led in addition to RGB
    KS12 = "KS12-"
    KS13 = "KS13-"

        
def gatt_from_prefix(prefix: DeviceNamePrefix) -> GattProfile | None:
    gatt_profile_by_prefix = {
        DeviceNamePrefix.KS01: GattProfile.new("ae00", "ae01", "ae02", "ae01"),
        DeviceNamePrefix.KS03: GattProfile.new("fff0", "fff3", "fff3", "fff3"),
        DeviceNamePrefix.KS03_New: GattProfile.new("afd0", "afd1", "afd2", "afd3"),
        DeviceNamePrefix.KS04: GattProfile.new("fff0", "fff3", "fff3", "fff3"),
        DeviceNamePrefix.KS04_New: GattProfile.new("ae00", "ae01", "ae02", "ae10"),
        DeviceNamePrefix.KS05: GattProfile.new_ex("ae00", "ae01", "ae00", "ae02", "ff00", "ff02"),
        DeviceNamePrefix.KS05: GattProfile.new("ae00", "ae01", "ae02", "ff02"),
        DeviceNamePrefix.KS07: GattProfile.new("ae00", "ae01", "ae02", "ae10"),
        DeviceNamePrefix.KS08: GattProfile.new("ae00", "ae01", "ae02", "ae10"),
        DeviceNamePrefix.KS09: GattProfile.new("ae00", "ae01", "ae02", "ae10"),
        DeviceNamePrefix.KS10: GattProfile.new("ae00", "ae01", "ae02", "ae10"),
        DeviceNamePrefix.KS11: GattProfile.new("ae00", "ae01", "ae02", "ae10"),
        DeviceNamePrefix.KS12: GattProfile.new("ae00", "ae01", "ae02", "ae10"),
        DeviceNamePrefix.KS13: GattProfile.new("ae00", "ae01", "ae02", "ae10"),
    }

    return gatt_profile_by_prefix.get(prefix)

def transmitter_from_prefix(prefix: DeviceNamePrefix, client: BleakClient) -> Transmitter | None:
    return BLETransmitter(client, gatt_from_prefix(prefix))

class Connection:
    def __init__(self, compiler: StateCompiler, transmitter: Transmitter):
        self._transmitter = transmitter
        self._compiler = compiler

    async def apply(self, state: LightState):
        await self._transmitter.send_all(
            self._compiler.compile(state)
        )

    async def disconnect(self):
        await self._transmitter.close()

@dataclass
class DeviceProfile:
    supported_commands: list[type[Command]]
    compiler: StateCompiler
    get_transmitter: Callable[[BleakClient], Transmitter] 

    async def connect(self, device: BLEDevice):
        # Instantiate device command compiler
        compiler = self.compiler()

        # Connect
        client = BleakClient(device)
        await client.connect()

        # Wrap BleakClient in a command transmitter
        transmitter = self.get_transmitter(client)

        return Connection(compiler, transmitter)

def make_transmitter_fetcher(prefix: DeviceNamePrefix):
    def fetcher(client: BleakClient):
        return transmitter_from_prefix(
            prefix,
            client
        )

    return fetcher

devices_by_prefix = {
        # DeviceNamePrefix.KS01: GattProfile.new("ae00", "ae01", "ae02", "ae01"),
        DeviceNamePrefix.KS03: DeviceProfile(
            supported_commands=[
                SwitchCommand,
                BrightnessCommand,
                RGBCommand,
                EffectCommand,
                SpeedCommand
            ],
            compiler=KS03OldCompiler,
            get_transmitter=make_transmitter_fetcher(DeviceNamePrefix.KS03),
        ),
        DeviceNamePrefix.KS03_New: DeviceProfile(
            supported_commands=[
                SwitchCommand,
                BrightnessCommand,
                RGBCommand,
                EffectCommand,
                SpeedCommand
            ],
            compiler=KS03NewCompiler,
            get_transmitter=make_transmitter_fetcher(DeviceNamePrefix.KS03_New),
        ),
        # DeviceNamePrefix.KS03_New: GattProfile.new("afd0", "afd1", "afd2", "afd3"),
        # # DeviceNamePrefix.KS04: GattProfile.new("fff0", "fff3", "fff3", "fff3"),
        # # DeviceNamePrefix.KS04_New: GattProfile.new("ae00", "ae01", "ae02", "ae10"),
        # # DeviceNamePrefix.KS05: GattProfile.new_ex("ae00", "ae01", "ae00", "ae02", "ff00", "ff02"),
        # # DeviceNamePrefix.KS05: GattProfile.new("ae00", "ae01", "ae02", "ff02"),
        # # DeviceNamePrefix.KS07: GattProfile.new("ae00", "ae01", "ae02", "ae10"),
        # # DeviceNamePrefix.KS08: GattProfile.new("ae00", "ae01", "ae02", "ae10"),
        # # DeviceNamePrefix.KS09: GattProfile.new("ae00", "ae01", "ae02", "ae10"),
        # # DeviceNamePrefix.KS10: GattProfile.new("ae00", "ae01", "ae02", "ae10"),
        # # DeviceNamePrefix.KS11: GattProfile.new("ae00", "ae01", "ae02", "ae10"),
        # # DeviceNamePrefix.KS12: GattProfile.new("ae00", "ae01", "ae02", "ae10"),
        # # DeviceNamePrefix.KS13: GattProfile.new("ae00", "ae01", "ae02", "ae10"),
    }

def device_from_prefix(prefix: DeviceNamePrefix | str) -> DeviceProfile | None:
    if isinstance(prefix, str):
        try:
            prefix = DeviceNamePrefix(prefix)
        except ValueError:
            return None

    return devices_by_prefix.get(prefix)

def device_profile_from_ble_device(bleak_device: BLEDevice) -> DeviceProfile | None:
# Fetch device based on first 5 characters of its name
    if bleak_device.name:
        return device_from_prefix(bleak_device.name[:5])

async def connect_to_ble_device(bleak_device: BLEDevice) -> Connection | None:
    # Fetch device based on first 5 characters of its name
    if device := device_profile_from_ble_device(bleak_device):
        return await device.connect(bleak_device)


from abc import ABC, abstractmethod

from cheshire.compiler.state import LightState
from cheshire.generic.platform_command import PlatformCommand


class StateCompiler(ABC):
    """Turns a light state into a list of platform commands which will achieve that state."""
    @abstractmethod
    def compile(self, state: LightState) -> list[PlatformCommand]:
        # Match on supported commands here. Ignore unsupported ones
        pass


from typing_extensions import TypedDict

from cheshire.generic.command import *

LightStateDict = TypedDict(
    'LightStateDict', 
    {
        'SwitchCommand': SwitchCommand,
        'BrightnessCommand': BrightnessCommand,
        'ColorTemperatureCommand': ColorTemperatureCommand,
        'RGBCommand': RGBCommand,
        'WhiteCommand': WhiteCommand,
        'EffectCommand': EffectCommand,
        'SpeedCommand': SpeedCommand
    },
    total=False
)

class LightState:
    """Represents the desired state of a light by collecting the latest of each 
    type of command."""
    def __init__(self) -> None:
        self._state: LightStateDict = {}

    def update(self, c: Command):
        self._state[c.get_type()] = c

    @property
    def state(self):
        return self._state 


from .constants import KS03NewScene
from cheshire.generic.platform_command import PlatformCommand
from struct import Struct

# TODO: Not all of these are KS03New
# Notably the warm white + color temp methods are KS11
# Move them to proper platform support modules

class KS03NewSwitchCommand(PlatformCommand):
    def __init__(self, on: bool):
        on_off = b"\xF0" if on else b"\x0F"
        self._bytes = b"\x5B" + on_off + b"\x00\xB5"

class KS03NewRGBWBrightnessSpeedCommand(PlatformCommand):
    _color_struct = Struct("BBBBBB")

    def __init__(self, red: int, green: int, blue: int, white: int, brightness: int, speed: int, is_rgb: bool):
        is_rgb_hex = b"\x01" if is_rgb else b"\x02"
        white = 0 if is_rgb else white
        self._bytes = b"\x5A\x00" + is_rgb_hex + self._color_struct.pack(red, green, blue, white, brightness, speed) + b"\xA5"

class KS03NewSceneCommand(PlatformCommand):
    _scene_struct = Struct("BBB")

    def __init__(self, scene_num: KS03NewScene, speed: int, brightness: int):
        self._bytes = b"\x5C\x00" + self._scene_struct.pack(scene_num + 128, speed, brightness) + b"\x00\xC5"

class KS03NewMusicModelCommand(PlatformCommand):
    _music_model_struct = Struct("B")
    _speed_struct = Struct("B")

    def __init__(self, music_model: int, speed: int):
        self._bytes = b"\x5A\x0A\xF0" + self._music_model_struct.pack(music_model) + self._speed_struct.pack(8 - speed) + b"\xA5"



class KS03NewLightColorCeilingRGB_CCTCommand(PlatformCommand):
    _color_struct = Struct("BBBBBB")

    def __init__(self, red: int, green: int, blue: int, warm: int, bright: int, speed: int):
        self._bytes = b"\x5A\x00\x00" + self._color_struct.pack(red, green, blue, warm, bright, speed) + b"\xA5"

class KS03NewMusicRBGBrightnessCommand(PlatformCommand):
    _RGB_struct = Struct("BBB")
    _brightness_struct = Struct("B")

    def __init__(self, red: int, green: int, blue:int, brightness: int):
        self._bytes = b"\x5A\x00\x01" + self._RGB_struct(red, green, blue) + b"\x00" + self._brightness_struct(brightness) + b"\x00\xA5"

class KS03NewMusicRGBWBrigtnessCommand(PlatformCommand):
    _music_RGBW_brightness_struct = Struct("BBBBB")

    def __init__(self, red: int, green: int, blue:int, white: int, brightness: int):
        self._bytes = b"\x5A\x00\x00" + self._music_RGBW_brightness_struct.pack(red, green, blue, white, brightness) + b"\x00\xA5"

class KS03NewWWCWCommand(PlatformCommand):
    _wwcw_struct = Struct("BBB")

    def __init__(self, ww: int, cw: int, brightness: int):
        self._bytes = b"\x5A\x02" + self._wwcw_struct.pack(ww, cw, brightness) + b"\x00\xA5"



from enum import IntEnum

class KS03OldScene(IntEnum):
    JUMP_7 = 0
    JUMP_3 = 1
    FADE_7 = 2
    FADE_3 = 3
    FLASH  = 4
    AUTO   = 5

class KS03OldMusicModel(IntEnum):
    # Slowly fades through colors. Noise increases speed.
    FADE_7_FAST_ON_NOISE = 0
    # Slowly pulses each color twice. Noise increases speed.
    TWO_FADE_FAST_ON_NOISE = 1
    # Solid color. Noise quickly cycles through colors.
    JUMP_ON_NOISE_PAUSE_QUIET = 2
    # Off. Noise turns on lights and quickly cycles through colors.
    JUMP_ON_NOISE_OFF_QUIET = 3




from dataclasses import dataclass

from .effect import Effect

@dataclass
class Command:
    """Generic meta-command which maps to one or more device commands."""
    def get_type(self):
        return type(self).__name__

@dataclass
class SwitchCommand(Command):
    on: bool

@dataclass
class BrightnessCommand(Command):
    brightness: int

@dataclass
class ColorTemperatureCommand(Command):
    ww: int
    cw: int

@dataclass
class RGBCommand(Command):
    red: int
    green: int
    blue: int

@dataclass
class WhiteCommand(Command):
    white: int

@dataclass
class EffectCommand(Command):
    effect: Effect

@dataclass
class SpeedCommand(Command):
    speed: int



from .constants import KS03OldScene
from cheshire.generic.platform_command import PlatformCommand
from struct import Struct

class KS03OldSwitchCommand(PlatformCommand):
    def __init__(self, on: bool):
        if on:
            self._bytes = b"\x7e\x04\x04\xf0\x01\x01\xff\x00\xef"
        else:
            self._bytes = b"\x7e\x04\x04\x10\x01\x00\xff\x00\xef"

class KS03OldBrightnessCommand(PlatformCommand):
    _brightness_struct = Struct("B")

    def __init__(self, brightness: int):
        self._bytes = b"\x7E\x04\x01" + self._brightness_struct.pack(brightness) + b"\xFF\xFF\xFF\x00\xEF"

class KS03OldRGBCommand(PlatformCommand):
    _rgb_struct = Struct("BBB")

    def __init__(self, red: int, green: int, blue: int):
        self._bytes = b"\x7E\x07\x05\x03" + self._rgb_struct.pack(red, green, blue) + b"\x00\xEF"
    
class KS03OldSpeedCommand(PlatformCommand):
    _speed_struct = Struct("B")

    def __init__(self, speed: int):
        self._bytes = b"\x7E\x04\x02" + self._speed_struct.pack(speed) + b"\xFF\xFF\xFF\x00\xEF"

class KS03OldSceneCommand(PlatformCommand):
    _scene_struct = Struct("B")

    def __init__(self, scene: KS03OldScene):
        self._bytes = b"\x7E\x05\x03" + self._scene_struct.pack(scene + 128) + b"\x03\xFF\xFF\x00\xEF"

class KS03OldMusicModelCommand(PlatformCommand):
    _music_model_struct = Struct("B")
    _speed_struct = Struct("B")

    def __init__(self, music_model: int, speed: int):
        self._bytes = b"\x7E\x0A" + self._music_model_struct.pack(music_model) + self._speed_struct.pack(8 - speed) + b"\xA5"


        # for kind, cmd in state.state.values():
        #     if isinstance(cmd, SwitchCommand):
        #         kstate.on = cmd.on
        #     if isinstance(cmd, BrightnessCommand):
        #         kstate.brightness
        #     if isinstance(cmd, RGBCommand):
        #         kstate.red = cmd.red
        #         kstate.green = cmd.green
        #         kstate.blue = cmd.blue
        #     if isinstance(cmd, EffectCommand):
        #         scene = effect_to_scene_mapping[cmd.effect]
        #         music_model= effect_to_music_model_mapping[cmd.effect]

        #         if scene:
        #             kstate.scene = scene.value
        #         if music_model
        #             kstate.music = music_modelvalue
        #     if isinstance(cmd, SpeedCommand):
        #         kstate.speed = cmd.speed



from cheshire.compiler.compiler import StateCompiler
from cheshire.compiler.state import LightState
from cheshire.generic.command import SwitchCommand, BrightnessCommand, RGBCommand, EffectCommand, SpeedCommand
from cheshire.generic.effect import Effect
from .platform_commands import *
from .constants import *

effect_to_scene_mapping = {
    # Effect.NONE: None,
    Effect.JUMP_7: KS03OldScene.JUMP_7,
    Effect.JUMP_3: KS03OldScene.JUMP_3,
    Effect.FADE_7: KS03OldScene.FADE_7,
    Effect.FADE_3: KS03OldScene.FADE_3,
    Effect.FLASH: KS03OldScene.FLASH,
    Effect.AUTO: KS03OldScene.AUTO,
}

effect_to_music_model_mapping = {
    # Effect.NONE: None,
    Effect.FADE_7_FAST_ON_NOISE: KS03OldMusicModel.FADE_7_FAST_ON_NOISE,
    Effect.TWO_FADE_FAST_ON_NOISE: KS03OldMusicModel.TWO_FADE_FAST_ON_NOISE,
    Effect.JUMP_ON_NOISE_PAUSE_QUIET: KS03OldMusicModel.JUMP_ON_NOISE_PAUSE_QUIET,
    Effect.JUMP_ON_NOISE_OFF_QUIET: KS03OldMusicModel.JUMP_ON_NOISE_OFF_QUIET,
}

class KS03OldCompiler(StateCompiler):
    def compile(self, state: LightState):
        platform_commands: list[PlatformCommand] = []

        # Set on/off state first
        if cmd := state.state.get('SwitchCommand'):
            platform_commands.append(
                KS03OldSwitchCommand(cmd.on)
            )

    
        # Set scene / music mode
        # (I think this needs to be set, then speed, then brightness
        # possibly with 100ms delays between commands)
        if cmd := state.state.get('EffectCommand'):
            if cmd.effect != Effect.NONE:
                scene = effect_to_scene_mapping.get(cmd.effect)
                music_model = effect_to_music_model_mapping.get(cmd.effect)
                # TODO support music mode
                if music_model is not None:
                    speed = 0
                    if speed_cmd := state.state.get('SpeedCommand'):
                        speed = speed_cmd.speed
                    platform_commands.append(
                        KS03OldMusicModelCommand(music_model, speed)
                    )
                if scene is not None:
                    platform_commands.append(
                        KS03OldSceneCommand(
                            scene
                        )
                    )

        # Set speed
        if cmd := state.state.get('SpeedCommand'):
            platform_commands.append(
                KS03OldSpeedCommand(
                    cmd.speed
                )
            )

        # Set brightness
        if cmd := state.state.get('BrightnessCommand'):
            platform_commands.append(
                KS03OldBrightnessCommand(int((cmd.brightness * 100) / 255))
            )
        
        # Set color
        if cmd := state.state.get('RGBCommand'):
            platform_commands.append(
                KS03OldRGBCommand(
                    int((cmd.red * 100) / 255),
                    int((cmd.green * 100) / 255),
                    int((cmd.blue * 100) / 255)
                )
            )

        return platform_commands



from dataclasses import dataclass

@dataclass
class PlatformCommand:
    """Low level command which maps directly to a device command"""
    def get_type(self):
        return type(self)

    def get_bytes(self) -> bytes:
        return self._bytes



from dataclasses import dataclass
from bleak import BleakClient
from bleak.backends.device import BLEDevice

from cheshire.communication.transmitter import Transmitter


@dataclass
class GattProfile:
    """
    Holds the BLE GATT service and characteristic paths to read write and notify
    """

    write_service: str
    write_characteristic: str
    notify_service: str
    notify_characteristic: str
    read_service: str
    read_characteristic: str

    @classmethod
    def new(cls, read_write_notify_service: str, write_characteristic: str, notify_characteristic: str, read_characteristic: str):
        return cls.new_ex(
            read_write_notify_service,
            write_characteristic,
            read_write_notify_service,
            notify_characteristic,
            read_write_notify_service,
            read_characteristic)

    @classmethod
    def new_ex(cls, write_service: str, write_characteristic: str, notify_service: str, notify_characteristic: str, read_service: str, read_characteristic: str) -> "GattProfile":
        return cls(
            write_service,
            write_characteristic,
            notify_service,
            notify_characteristic,
            read_service,
            read_characteristic)


class BLETransmitter(Transmitter):
    def __init__(self, client: BleakClient, gatt: GattProfile):
        self._client = client
        self._gatt = gatt
        self._characteristic = None
        for service in self._client.services:
            # Match the short Bluetooth UUID
            if service.uuid[4:8] == self._gatt.write_service:
                for characteristic in service.characteristics:
                    if characteristic.uuid[4:8] == self._gatt.write_characteristic:
                        self._characteristic = characteristic
                        return
        raise ConnectionError(f"Expected BLEClient to support a characteristic with UUID[4:8] == {self._gatt.write_characteristic}")

    async def close(self):
        await self.disconnect()

    async def disconnect(self):
        await self._client.disconnect()

    async def send_raw(self, raw_command: bytes):
        return await self._client.write_gatt_char(self._characteristic, raw_command, False)


from abc import ABC, abstractmethod

from cheshire.generic.platform_command import PlatformCommand


class Transmitter(ABC):
    @abstractmethod
    async def close(self):
        """Must be called before transmitter is discarded so it can close connections etc."""
        pass

    @abstractmethod
    async def send_raw(self, raw_cmd: bytes):
        pass

    async def send(self, cmd: PlatformCommand):
        return await self.send_raw(cmd.get_bytes())
    
    async def send_all(self, cmds: list[PlatformCommand]):
        for cmd in cmds:
            await self.send(cmd)




from cheshire.compiler.compiler import StateCompiler
from cheshire.compiler.state import LightState
from cheshire.generic.command import *
from cheshire.generic.effect import Effect
from .constants import *
from .platform_commands import *


effect_to_scene_mapping = {
    # Effect.NONE: None,
    Effect.JUMP_7: KS03NewScene.JUMP_7,
    Effect.JUMP_3: KS03NewScene.JUMP_3,
    Effect.FADE_7: KS03NewScene.FADE_7,
    Effect.FADE_3: KS03NewScene.FADE_3,
    Effect.FLASH: KS03NewScene.FLASH,
    Effect.AUTO: KS03NewScene.AUTO,
}

effect_to_music_model_mapping = {
    # Effect.NONE: None,
    Effect.FADE_7_FAST_ON_NOISE: KS03NewMusicModel.FADE_7_FAST_ON_NOISE,
    Effect.TWO_FADE_FAST_ON_NOISE: KS03NewMusicModel.TWO_FADE_FAST_ON_NOISE,
    Effect.JUMP_ON_NOISE_PAUSE_QUIET: KS03NewMusicModel.JUMP_ON_NOISE_PAUSE_QUIET,
    Effect.JUMP_ON_NOISE_OFF_QUIET: KS03NewMusicModel.JUMP_ON_NOISE_OFF_QUIET,
}

class KS03NewCompiler(StateCompiler):
    def compile(self, state: LightState):
        platform_commands: list[PlatformCommand] = []

        # Set on/off state first
        if cmd := state.state.get('SwitchCommand'):
            platform_commands.append(
                KS03NewSwitchCommand(cmd.on)
            )

        # Extract brightness used in other commands
        brightness = 254
        if cmd := state.state.get('BrightnessCommand'):
            brightness = cmd.brightness

        # Extract speed used in other commands
        speed = 0
        if speed_cmd := state.state.get('SpeedCommand'):
            speed = speed_cmd.speed
        
        # KS03New uses different commands for setting color brightness etc.
        # for music and scene modes than in simple light mode
        if cmd := state.state.get('EffectCommand'):
            if cmd.effect != Effect.NONE:
                music_model= effect_to_music_model_mapping.get(cmd.effect)
                scene = effect_to_scene_mapping.get(cmd.effect)

                if music_model:
                    platform_commands.append(
                        KS03NewMusicModelCommand(music_model, speed)
                    )
                    
                if scene:
                    platform_commands.append(
                        KS03NewSceneCommand(scene, speed, brightness)
                    )

                return
                
        
        r = 100
        g = 100
        b = 100

        if rgb_cmd := state.state.get('RGBCommand'):
            r = int((rgb_cmd.red * 100) / 255)
            g = int((rgb_cmd.green * 100) / 255)
            b = int((rgb_cmd.blue * 100) / 255)

        platform_commands.append(
            KS03NewRGBWBrightnessSpeedCommand(
                red=r, 
                green=g,
                blue=b,
                white=0, # KS03New doesn't support warm white
                brightness=brightness,
                speed=speed,
                is_rgb=True # Double-disables warm white :)
            )
        )

        return platform_commands